#include <linux/init.h>
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/skbuff.h>
#include <linux/netfilter.h>
#include <linux/netfilter_ipv4.h>
#include <linux/ip.h>
#include <linux/tcp.h>
#include <linux/udp.h>

#include <net/ip.h>
#include <net/tcp.h>
#include <net/udp.h>
#include <net/route.h>
#include <net/sock.h>


#include "ttm.h"


struct ttm_peer {
    __u16     af;
    __be16  port;
    union {
        __u32       all[4];
        __be32      ip;
        __be32      ip6[4];
        struct in_addr  in;
        struct in6_addr in6;
    };
};

#define TTM_BASE_CTL    (64+1024+64+64+64+64)  /* base */
#define TTM_SO_GET_PEER (TTM_BASE_CTL+1)
#define TTM_SO_GET_MAX  (TTM_SO_GET_PEER)

#define GET_CMDID(cmd)      (cmd - TTM_BASE_CTL)
#define GET_TTM_INFO_LEN    (sizeof(struct ttm_peer))

static const unsigned char get_arglen[GET_CMDID(TTM_SO_GET_MAX) + 1] = {
    [GET_CMDID(TTM_SO_GET_PEER)] = GET_TTM_INFO_LEN,
};

static int
do_ttm_get_info(struct sock *sk, int cmd, void __user * user, int *len)
{
    struct ttm_peer peer = { 0 };
    struct ttm_data_v6 *data6;
    struct ttm_data *data;
    int ret = 0;

    if (*len < get_arglen[GET_CMDID(cmd)]) {
        printk("get_ctl: len %u < %u\n",
               *len, get_arglen[GET_CMDID(cmd)]);
        return -EINVAL;
    }

    switch (cmd) {
    case TTM_SO_GET_PEER:
		data6 = (struct ttm_data_v6 *)sk->sk_user_data;
		data = (struct ttm_data *)(&sk->sk_user_data);
	
		/* 4to4 case */
		/* notice that data maybe **ptr of data_v64 here */
		if (NULL != sk->sk_user_data && 
			data->opsize == TCPOLEN_TTM &&
			(data->opcode == TCPOPT_TTM_CIP ||
			data->opcode == TCPOPT_FNAT)) {
			peer.af = AF_INET;
			peer.port = data->port;
			peer.ip = data->ip; 
		}		
		/* 6to4 case */
		else if (data6->opcode == TCPOPT_TTM_CIP_V6 &&
			data6->opsize == TCPOLEN_TTM_v6) {
			peer.af = AF_INET6;
			peer.port = data6->port;
			memcpy(peer.all, data6->ipv6, sizeof(data6->ipv6));
		}
		/* normal case */
		else {
            struct inet_sock *inet  = inet_sk(sk);
            peer.af = AF_INET;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,33)
            peer.port = inet->inet_dport;
            peer.ip = inet->inet_daddr;
#else
            peer.port = inet->dport;
            peer.ip = inet->daddr;
#endif
        }

        if (copy_to_user(user, &peer, GET_TTM_INFO_LEN) != 0)
            ret = -EFAULT;

        break;
    default:
        ret = -EINVAL;
    }
	
    return ret;
}


static struct nf_sockopt_ops ttm_sockopts = {
    .pf = PF_INET,
    .get_optmin = TTM_BASE_CTL,
    .get_optmax = TTM_SO_GET_MAX + 1,
    .get = do_ttm_get_info,
    .owner = THIS_MODULE,
};

int
ttm_init_ctl(void)
{
    int ret;


    ret = nf_register_sockopt(&ttm_sockopts);
    if (ret) {
        printk("can't register ttm sockopt.\n");
    }
    return ret;
}

void
ttm_uninit_ctl(void)
{
    nf_unregister_sockopt(&ttm_sockopts);
}


EXPORT_SYMBOL(ttm_init_ctl);
EXPORT_SYMBOL(ttm_uninit_ctl);

